"""
Connexion network infrastructure

It consists of three major parts:
	* UDP client/server socket for inter-cx communications
	* ZeroConf mDNS engine
	* OS network control subsystem: for Linux, it is [rt]netlink

"""

# 	Copyright (c) 2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from ctypes import *
from socket import AF_INET, SOL_SOCKET, SO_REUSEADDR, SOCK_DGRAM, socket, inet_ntoa, inet_aton
from Crypto.Cipher import AES
from Crypto.Util.randpool import RandomPool
from pickle import dumps, loads
from threading import Timer
from sys import version_info

import traceback
import time

from cxnet.common import hline
from cxutil.utils import opts
from cxutil.UID import UID
from cxnet import zeroconf

from manager import ManagerThread,push,passive
from logger import log

CL_PORT = 40323
CL_DOMAIN = "_cx._udp.local"

class mDNSlistener(object,log):
	'''
	mDNS cache updater
	'''

	def __init__(self,server,bus):
		self.bus = bus
		self.server = server

	def removeService(self, zeroconf, type, name):
		self.log("info","unregistered service `%s'" % (name))

	def addService(self, zeroconf, type, name):
		info = self.server.getServiceInfo(type, name)
		self.log("info","registered service `%s'" % (name))


class mDNSannouncer(zeroconf.Announcer,log):
	'''
	mDNS cache announcer
	'''
	def __init__(self,bus):
		self.bus = bus
	
	def signal(self,sig,action,record):
		try:
			
			data = opts({
				"action": action,
				"host": record.name,
			})


			if (record.clazz == zeroconf._CLASS_IN) and (record.type == zeroconf._TYPE_A):
				data["address"] = inet_ntoa(record.address)
			elif (record.clazz == zeroconf._CLASS_IN) and (record.type == zeroconf._TYPE_SRV):
				data["port"] = record.port

			elif (record.clazz == zeroconf._CLASS_IN) and (record.type == zeroconf._TYPE_TXT):
				svc = zeroconf.ServiceInfo(record.name, record.name)
				svc.setText(record.text)
				data["properties"] = opts(svc.getProperties())
				
				# FIXME:
				# it's a dirty hack, please isolate this code
				try:
					self.bus.put(
						"lockd",
						opts({
							"map": "signal_%s" % (data["properties"]["role"]),
							"action": action,
							"domain": record.name[record.name.index(".") + 1:],
							"alias": data["properties"]["alias"],
						})
					)

				except Exception,e:
					print e
					self.log("error",traceback.format_exc())
			
			self.bus.put(
				"state",
				{
					"map": "signal",
					"signal": sig,
					"data": data,
				}
			)
		except:
			self.log("error","signal exception: %s" % (str(traceback.format_exc())))
	
	def add(self,record):
		self.signal(("mdns.cache",),"add",record)
		self.log("info","cache add: %s" % (repr(record)))

	def remove(self,record):
		self.signal(("mdns.cache",),"remove",record)
		self.log("info","cache remove: %s" % (repr(record)))
	
	def expire(self,record):
		self.signal(("mdns.cache",),"expire",record)
		self.log("info","cache expire: %s" % (repr(record)))


## Common exceptions
class mDNS_AlreadyRegistered(Exception):
	def __str__(self):
		return "Domain browser for `%s` already registered" % (self.message)


## Lookup exceptions
class mDNS_ServiceNotFound(Exception):
	def __str__(self):
		return "Service not found"

class mDNS_HostNotFound(Exception):
	def __str__(self):
		return "Host not found"

class mDNS_PortNotFound(Exception):
	def __str__(self):
		return "Port not found"

class mDNS_TextNotFound(Exception):
	def __str__(self):
		return "Text not found"



class mDNSengine(object,log):
	'''
	ZeroConf mDNS object with browsing support
	'''

	def __init__(self, address, join, bus, adaptive=True, heartbeat=True):
		# set up ZeroConf mDNS
		self.bus = bus
		self.server = zeroconf.Zeroconf(address, join, adaptive, heartbeat=heartbeat)
		self.announcer = mDNSannouncer(bus)
		self.server.addCacheHook(self.announcer)
		self.listener = mDNSlistener(self.server, self.bus)
		self.browsers = {}

	def registerService(self,name,address="0.0.0.0",port=CL_PORT,type=CL_DOMAIN,properties={},records=[zeroconf._TYPE_A, zeroconf._TYPE_SRV, zeroconf._TYPE_TXT],ttl=zeroconf._DNS_TTL):
		# register the service and return a reference
		if address:
			address = inet_aton(address)
		svc = zeroconf.ServiceInfo(
			type='%s' % (type),
			name='%s.%s' % (name, type),
			address=address,
			port=port,
			weight=0,
			priority=0,
			properties=properties,
			records=records,
			ttl=ttl
		)
		self.server.registerService( svc )
		# FIXME: return weak reference
		return svc
	
	def registerDomain(self,domain):
		# set up mDNS browser
		if self.browsers.has_key(domain):
			raise mDNS_AlreadyRegistered(domain)
		else:
			self.browsers[domain] = zeroconf.ServiceBrowser(self.server, "%s" % (domain), self.listener)

	def lookupPTR(self,domain):
		"""
		Lookup a mDNS PTR record for a domain
		"""
		result = []
		for i in self.server.cache.entriesWithName(domain):
			if (i.clazz == zeroconf._CLASS_IN) and (i.type == zeroconf._TYPE_PTR):
				result.append(i.alias)
		
		return result

	def lookup(self,host,domain=None,types=[zeroconf._TYPE_A, zeroconf._TYPE_SRV],nowait=False):
		"""
		Lookup a mDNS entry for address, port or text
		"""

		if not domain:
			domain = host[host.index(".") + 1:]

		if not self.server.cache.entriesWithName(host):
			if nowait:
				raise mDNS_ServiceNotFound()
			else:
				self.server.getServiceInfo(domain, host)

		svc = zeroconf.ServiceInfo(domain, host)

		for i in self.server.cache.entriesWithName(host):
			if (i.clazz == zeroconf._CLASS_IN) and (i.type == zeroconf._TYPE_A):
				svc.address = inet_ntoa(i.address)
			elif (i.clazz == zeroconf._CLASS_IN) and (i.type == zeroconf._TYPE_SRV):
				svc.port = i.port
			elif (i.clazz == zeroconf._CLASS_IN) and (i.type == zeroconf._TYPE_TXT):
				svc.setText(i.text)

		if (not svc.address) and (zeroconf._TYPE_A in types):
			raise mDNS_HostNotFound()
		
		if (not svc.port) and (zeroconf._TYPE_SRV in types):
			raise mDNS_PortNotFound()
		
		if (not svc.text) and (zeroconf._TYPE_TXT in types):
			raise mDNS_TextNotFound()

		return ((svc.getAddress(),svc.getPort()),svc.getProperties())


###
#
# Connexion protocol definition (FIXME: one should extract
# this out to cxnet lib)
#
# 8<---------------------------------------------------------------------
#
# Temporary protocol structure, should be revised to include nonce
#
# +--------+--------+--------+--------+
# |              msg_len              |
# +--------+--------+--------+--------+
# |     version     |      flags      |
# +--------+--------+--------+--------+
# |           ... data ...            |
# +-----------------------------------+
#
class cxhdr(BigEndianStructure):
	'''
	'''
	# BE == network byte order, so, we do not need
	# hton[sl]() and ntoh[sl]()
	_fields_ = [
		("msg_len",	c_int32),
		("version",	c_int16),
		("flags",	c_int16),
	]


CX_MAX_LENGTH = 4096
CX_MAX_MSG = CX_MAX_LENGTH - sizeof(cxhdr)

class cxmsg(Structure):
	'''
	'''
	_fields_ = [
		("header",	cxhdr),
		("data",	c_byte * CX_MAX_MSG),
	]
# 8<---------------------------------------------------------------------
#
###

###
#
# packet debug hooks
def skip(*argv):
	pass

def dump(bus,packet,size,line="packet"):
	bus.put("logger",("debug", "%s:\n%s" % (line,hline(packet,size))))

class cl_socket(object, log):
	'''
	UDP client/server socket for inter-cx communications
	
	Provides [un]reliable encrypted transport for inter-cx communications
	'''
	
	sock = None
	fd = None

	key = None
	cipher = None
	block = None
	rand = None

	recv_buffer = None
	send_buffer = None

	bus = None

	def __init__(self,address,port,bus,key=None,keyfile=None,block=16,debug=False):
		'''
		address		- IP address to bind to
		port		- UDP port to use
		bus 		- internal bus to send debug to
		
		key	or
		keyfile		- initialize cipher key from string or file
		block		- cipher block size
		
		debug		- turn packet logging on/off
		'''
		
		if (not key) and (not keyfile):
			raise Exception("specify encryption key or keyfile")

		self.block = block

		if keyfile:
			f = open(keyfile,'r')
			self.key = f.read(self.block)
		else:
			self.key = key

		if len(self.key) < self.block:
			raise Exception("not enough symbols in the key")

		# socket operations
		self.sock = socket(AF_INET,SOCK_DGRAM)
		self.sock.setsockopt(SOL_SOCKET,SO_REUSEADDR,1)

		self.sock.bind((address,port))
		self.fd = self.sock.fileno()
		
		# buffers & headers
		self.recv_buffer = create_string_buffer(CX_MAX_LENGTH)
		self.send_buffer = create_string_buffer(CX_MAX_LENGTH)
		
		# encryption
		self.cipher = AES.new(self.key,AES.MODE_CBC)
		self.rand = RandomPool(CX_MAX_LENGTH)
		self.randomize()

		# internal bus, send only
		self.bus = bus

		# debug hook
		if debug:
			self.hook = dump
		else:
			self.hook = skip

	def recv(self):
		'''
		Receive and decrypt a packet
		'''
		try:
			# receive
			# python >= 2.5
			try:
				(bytes,(address,port)) = self.sock.recvfrom_into(self.recv_buffer)
			# python < 2.5
			except:
				if version_info[:2] < (2,5):
					(s,(address,port)) = self.sock.recvfrom(CX_MAX_LENGTH)
					x = create_string_buffer(s)
					memmove(self.recv_buffer,x,sizeof(x))
				else:
					raise
			# parse header
			header = cxhdr.from_address(addressof(self.recv_buffer))
			# get plaintext message and load it

			self.hook(self.bus, self.recv_buffer, header.msg_len + sizeof(header), "cl_socket.recv()")

			n = string_at( addressof(self.recv_buffer) + sizeof(header), header.msg_len )
			m = self.decrypt(n)
			msg = loads(m)
			# localize recipient (FIXME: no forwarding yet)
			msg["to"] = msg["to"].split("@")[0]

			self.log("debug","<< %s" % (msg))
			return msg
		except Exception,e:

			self.log("error","cl_socket.recv(): %s" % (e))
			self.log("debug","%s" % (traceback.format_exc()))
			return None

	def send(self,host,packet):
		'''
		Encrypt and send a packet
		'''
		self.log("debug", ">> %s" % (packet))
		try:
			# create a message
			msg = cxmsg()
			# encrypt payload
				# get an encrypted string
			c = self.encrypt(dumps(packet))
				# create ctypes string buffer from this string
			x = create_string_buffer(c,len(c))
				# copy the buffer to the packet
			memmove(addressof(msg) + sizeof(msg.header),addressof(x),len(c))
			# NOTE: I don't know, why c_char * CX_MAX_MSG does not work.
			# FIXME: fix it later...

			# construct header
			msg.header.msg_len = len(c)
			msg.header.version = 1
			msg.header.flags = 0

			self.hook(self.bus, msg, msg.header.msg_len + sizeof(msg.header), "cl_socket.send()")

			# send a packet
			self.sock.sendto(string_at(addressof(msg),msg.header.msg_len + sizeof(msg.header)),host)

		except Exception,e:
			self.log("error","cl_socket.send(): %s" % (e))
			self.log("debug","%s" % (traceback.format_exc()))
			return

	def decrypt(self,data):
		'''
		Decrypt data
		'''
		# strip IV
		return self.cipher.decrypt(data)[self.block:]

	def encrypt(self,data):
		'''
		Encrypt data with block cipher
		'''
		# create IV
		self.randomize()
		data = self.rand.get_bytes(self.block) + data

		# calculate alignment
		a = ((len(data) + self.block - 1) & ~ (self.block - 1)) - len(data)

		# justify data with random string
		self.randomize()
		data += self.rand.get_bytes(a)
		self.randomize()

		# encrypt it
		return self.cipher.encrypt(data)


	def randomize(self):
		'''
		Update random pool
		'''
		while self.rand.entropy < CX_MAX_LENGTH:
			self.rand.add_event()


class CXmutexCreateException(Exception):
	def __str__(self):
		return "cannot create distributed mutex: %s" % (self.message)

class CX_NoLeader(Exception):
	def __str__(self):
		return "no leader for domain %s" % (self.message)

class CX_MultipleLeaders(Exception):
	def __str__(self):
		return "multiple leaders for domain %s" % (self.message)


class cl_lockd(ManagerThread):
	'''
	Connexion locking daemon
	
	'''

	passive_generators = []
	
	def __init__(self,bus,mdns,alias):

		ManagerThread.__init__(self,bus,"lockd")
		
		self.mdns = mdns
		self.alias = alias
		self.ds = {}

	# checked
	def init(self,argv):
		
		mutex = argv["mutex"]

		domain = mutex.domain
		mutex.uid = UID()
		
		###
		#
		# Register mutex domain, if not exists
		#
		###
		if not self.ds.has_key(domain):
			# for each mutex domain there is a structure that holds info about:
			self.ds[domain] = opts({
				"uid": mutex.uid,	# uuid of the service
				"svc": None,		# ServiceInfo reference
				"queue": [],		# queue of uninitialized mutexes, without leader
				"counter": 0,		# election counter
				"queues": {},		# mutex queues, each queue is a list of tuples (node address, service ID)
				"watchdog": None	# election watchdog
			})
		
			try:
				# register service
				self.ds[domain]["svc"] = self.mdns.registerService(
					name = mutex.uid,
					type = domain,
					records = [zeroconf._TYPE_TXT],
					properties = {
						"role": "candidate",
						"alias": self.alias,
					},
					ttl = 2 # zeroconf._DNS_TTL
				)
			except Exception,e:
				self.log("error","domain registration error: `%s`" % (e))
				return e

	# checked
	def getRing(self,domain):
		'''
		Get all host names for a zone (domain)
		'''
		ring = self.mdns.lookupPTR(domain)
		ring.sort()
		return ring

	# checked
	def getLeader(self,domain):
		'''
		Get a leader for a zone (domain), or raise an exception:
		
		* CX_MultipleLeaders
		* CX_NoLeader
		'''
		
		ring = self.getRing(domain)
		
		# lookup a leader
		leader = None
		for i in ring:
			try:
				((address,port),properties) = self.mdns.lookup(i,domain=domain,types=[zeroconf._TYPE_TXT],nowait=True)

				if (properties["role"] == "leader"):
					if leader:
						# FIXME: log possible attack
						self.log("warning","multiple leaders: possible attack")
						raise CX_MultipleLeaders(domain)
					else:
						leader = properties["alias"]
			except CX_MultipleLeaders:
				raise
			except:
				pass

		if not leader:
			raise CX_NoLeader(domain)

		return leader

	# checked
	def assertMutex(self,domain,name,sid):
		while True:
			try:
				if self.ds[domain].queues[name][0][1] == sid:
					yield self.ds[domain].queues[name]
					return
			except:
				pass
			
			yield None

	# checked
	def assertLeader(self,domain):
		while True:
			try:
				yield self.getLeader(domain)
				return
			except CX_NoLeader:
				self.runElection(domain)
			except:
				pass
			yield None

	@passive(passive_generators)
	def acquire(self,argv):

	
		domain = argv["mutex"].domain
		name = argv["mutex"].name
		sid = "%s.%s" % (self.ds[domain]["uid"], domain)

		for leader in self.assertLeader(domain):
			if not leader: yield None
		self.log("debug","leader asserted: %s" % (leader))

		argv["watchdog"] = True
		if not self.acquire_init(argv,5):
			argv["watchdog"] = False

		for mutex in self.assertMutex(domain,name,sid):
			if not mutex: yield None
		self.log("debug","mutex asserted: %s" % (mutex))
		
		if argv["watchdog"]:
			try:
				argv["watchdog"].cancel()
			except:
				pass
		else:
			# we can reach this code only after mutex acquisition timeout
			# so, return False
			self.log("error","cannot acquire mutex `%s`" % (mutex))
			yield False
			return

		yield True

	def acquire_init(self,argv,countdown = 0):

		domain = argv["mutex"].domain
		name = argv["mutex"].name
		sid = "%s.%s" % (self.ds[domain]["uid"], domain)

		# try to find mutex in the list
		try:
			if sid in map(lambda x: x[1], self.ds[domain].queues[name]):
				self.log("debug","mutex already queued, request/watchdog aborted")
				argv["watchdog"] = True
				return False
		except:
			pass

		if countdown > 0:
			argv["watchdog"] = Timer(5,self.acquire_init,[argv,countdown - 1])
			argv["watchdog"].start()
			self.log("debug","mutex acquire watchdog started")
		else:
			self.log("critical","timed out, watchdog aborted")
			argv["watchdog"] = False

		try:
			self.bus.put(
				"lockd@%s" % (self.getLeader(domain)),
				{
					"map": "cl_acquire",
					"name": name,
					"domain": domain,
					"sid": sid,
				}
			)
		except CX_NoLeader:
			self.runElection(domain)
		except:
			pass
	
		return True


	@passive(passive_generators)
	def release(self,argv):
		
		leader = None
		domain = argv["mutex"].domain

		for leader in self.assertLeader(domain):
			if not leader: yield None

		self.log("debug","leader asserted: %s" % (leader))

		self.release_init(argv,5)



	def release_init(self,argv,countdown = 0):
		
		domain = argv["mutex"].domain
		name = argv["mutex"].name
		sid = "%s.%s" % (self.ds[domain]["uid"], domain)

		# try to find mutex in the list
		try:
			if sid not in map(lambda x: x[1], self.ds[domain].queues[name]):
				self.log("debug","mutex dequeued, request/watchdog aborted")
				argv["watchdog"] = True
				return False
		except:
			pass

		if countdown > 0:
			argv["watchdog"] = Timer(5,self.release_init,[argv,countdown - 1])
			argv["watchdog"].start()
			self.log("debug","mutex release watchdog started")
		else:
			self.log("critical","timed out, watchdog aborted")
			argv["watchdog"] = False

		try:
			self.bus.put(
				"lockd@%s" % (self.getLeader(domain)),
				{
					"map": "cl_release",
					"name": name,
					"domain": domain,
					"sid": sid
				}
			)
		except CX_NoLeader:
			self.runElection(domain)
		except:
			pass
	
		return True


	def cl_update(self,argv):
		self.ds[argv["domain"]]["queues"][argv["name"]] = argv["queue"]
		self.cl_check_dead(argv["domain"],argv["name"],argv["queue"])

	def cl_announce(self,domain,name,q):
		for i in self.getRing(domain):
			try:
				((address,port),properties) = self.mdns.lookup(i,domain=domain,types=[zeroconf._TYPE_TXT],nowait=True)
			
				self.bus.put(
					"lockd@%s" % (properties["alias"]),
					{
						"map": "cl_update",
						"domain": domain,
						"name": name,
						"queue": q
					}
				)
			except:
				pass

	# checked
	def cl_check_dead(self,domain,name,q):
		# validate address
		while q:
			_sid = q[0][1]
			try:
				# get service record from DNS for this alias
				for i in self.getRing(domain):
					if i == _sid:
						self.log("debug","sid %s validated: found a live mutex lock" % (_sid))
						break
				else:
					raise Exception("sid %s seems to be dead" % (_sid))
					
				# so far we have a correct lock in the queue, abort
				break
			except:
				# dead mutex, clean up

				# silently drop what release returns, because
				# mutex owner is dead and there is no need to
				# alert him
				self.log("info","drop dead mutex lock for %s / %s" % q[0])
				self.cl_release({
					"domain": domain,
					"name": name,
					"addr": q[0][0],
					"sid": _sid,
				})

		if q:
			return q[0]

	def cl_acquire(self,argv):
		self.log("debug","acquire request for `%s` from `%s`" % (argv["name"], argv["addr"]))
		qs = self.ds[argv["domain"]]["queues"]
		rq = None

		try:
			q = qs[argv["name"]]
		except:
			q = qs[argv["name"]] = []
		
		if q:
			rq = self.cl_check_dead(argv["domain"], argv["name"], q)

		# one should queue mutex only if there is no such sid in the queue:
		# we can receive duplicates
		if argv["sid"] not in map(lambda x: x[1], q):
			q.append((argv["addr"],argv["sid"]))
		else:
			self.log("warning","double lock attempt for `%s`" % (argv))

		self.cl_announce(argv["domain"],argv["name"],q)

	# checked
	def cl_release(self,argv):
		qs = self.ds[argv["domain"]]["queues"]
		try:
			q = qs[argv["name"]]
		except:
			self.log("warning","release request for non-existent mutex `%s` from node `%s`" % (argv["name"], argv["addr"]))
			return

		i = q[0][1]
		if argv["sid"] != i:
			self.log("warning","release for mutex `%s` from non-owner `%s`" % (argv["name"], argv["addr"]))
			return

		q.pop(0)
		self.cl_announce(argv["domain"],argv["name"],q)

		self.log("debug","release request for `%s` from `%s`" % (argv["name"], argv["addr"]))

	def watchElection(self,domain):
		self.ds[domain].watchdog = None
		try:
			self.getLeader(domain)
		except CX_NoLeader:
			self.forceElection(domain)

	def forceElection(self,domain):
		
		# start election algorithm
		self.log("debug","election started for `%s`" % (domain))

		self.ds[domain]["svc"].setProperty("role","candidate")
		self.ds[domain]["counter"] += 1
		if not self.ds[domain]["watchdog"]:
			self.ds[domain]["watchdog"] = Timer(5,self.watchElection,[domain])
			self.ds[domain]["watchdog"].start()
			self.log("debug","election watchdog started for `%s`" % (domain))

		# create a token
		token = {
			"map": "relayElection",
			"uid": self.ds[domain]["uid"],
			"domain": domain,
			"counter": self.ds[domain]["counter"]
		}
		self.sendNext(domain,token)

	# checked
	def runElection(self,domain):
	
		if self.ds[domain]["counter"]:
			self.log("debug","election already rinning")
			return
		
		self.forceElection(domain)

	# checked
	def relayElection(self,argv):
		domain = argv["domain"]
		uid = argv["uid"]
		counter = argv["counter"]
		live = False
		node = ""

		# got to relay election
		
		euid = self.ds[domain]["uid"]

		# 1. check for dead mutexes:
		for (i,k) in self.ds[domain].queues.items():
			# i -- mutex name
			# k -- mutex queue
			self.cl_check_dead(domain,i,k)

		# 2. send our token (if we are not in the process)
		if self.ds[domain]["counter"] < counter:
			self.ds[domain]["counter"] = counter
			token = {
				"map": "relayElection",
				"uid": euid,
				"domain": domain,
				"counter": self.ds[domain]["counter"]
			}
			self.sendNext(domain,token)

		# 3. analyze received token
		if uid == euid:
			# our token or it's our mutex in the head: we're leader
			self.ds[domain]["svc"].setProperty("role","leader")
			self.ds[domain]["svc"].ttl = 2

		elif uid < euid:
			# respect priority
			self.sendNext(domain,argv)
			self.ds[domain]["svc"].setProperty("role","candidate")
			
			# The scheme with only one beacon in the network
			# requires leadership migration with mutex acquisition.
			# 
			# While we're not ready to make this great step for
			# mankind, use _all_ nodes for beacons
			self.ds[domain]["svc"].ttl = 2 # zeroconf._DNS_TTL

		else:
			# just drop the token
			pass

	# checked
	def sendNext(self,domain,token):
		while True:
			ring = self.getRing(domain)
			ring.sort()
			try:
				idx = ring.index("%s.%s" % (self.ds[domain]["uid"], domain))
			except:
				self.log("critical", traceback.format_exc())
				return
			
			if idx > 0:
				idx -= 1
			else:
				idx = len(ring) - 1

			next = ring[idx]
			try:
				((address,port),properties) = self.mdns.lookup(next,domain=domain,types=[zeroconf._TYPE_TXT],nowait=True)
				self.bus.put("lockd@%s" % (properties["alias"]), token)
				return
			except Exception,e:
				self.log("debug","lookup error %s for host `%s` domain `%s`" % (e,next,domain))

	def signal_candidate(self,argv):

		# check for dead mutexes:
		for (i,k) in self.ds[argv.domain].queues.items():
			# i -- mutex name
			# k -- mutex queue
			self.cl_check_dead(argv.domain,i,k)

		try:
			if (argv.action == "add") and (argv.alias != self.alias):
				# replicate mutex table for the domain
				for (i,k) in self.ds[argv.domain].queues.items():
					# i -- mutex name
					# k -- mutex queue
					self.bus.put(
						"lockd@%s" % (argv.alias),
						{
							"map": "cl_update",
							"domain": argv.domain,
							"name": i,
							"queue": k
						}
					)

		except Exception,e:
			self.log("error","candidates update error: %s" % (e))

	def signal_leader(self,argv):

		# FIXME: get this out to a separate method

		# check for dead mutexes:
		for (i,k) in self.ds[argv.domain].queues.items():
			# i -- mutex name
			# k -- mutex queue
			self.cl_check_dead(argv.domain,i,k)

		if argv.action == "add":
			self.ds[argv.domain].counter = 0
			try:
				self.ds[argv.domain].watchdog.cancel()
				self.ds[argv.domain].watchdog = None
				self.log("debug","election watchdog for `%s` aborted" % (argv.domain))
			except:
				pass
		else:
			self.runElection(argv.domain)



class cl_mutex(object):
	
	def __init__(self,bus,name,domain):
		
		self.lockd = "lockd"
		self.name = name
		self.domain = domain
		self.bus = bus.put2("dispatcher2","connect")
		self.bus.put(self.lockd, {"map": "init", "mutex": self})
	
	def acquire(self):
		return self.bus.put2(self.lockd, {"map": "acquire", "mutex": self})
	
	def release(self):
		self.bus.put(self.lockd, {"map": "release", "mutex": self})
	

class net_setup(object):
	'''
	Abstracts network control procedures like 'get interface', 'set route'
	and so on, and so forth.
	'''
	pass
